#pragma check_stack(off)
#include <dos.h>
#include <string.h>

#include "rmount.h"
#include "ems.h"
#include "debug.h"
#include "lzrw1.h"
#include "xdr.h"
#include "pport.h"
#include "syscall.h"

#define MIN_COMPRESS 	  100u	/* min size of buffer to be compressed */

#define MAPBUFFER	emsMapPage(emsCompression, 0, 0);\
					emsMapPage(emsCompression, 1, 1);\
					emsMapPage(emsCompression, 2, 2);\
					emsMapPage(emsCompression, 3, 3);

extern uint8 		use_compress;
extern emshandle_t	emsCompression;

uint16 sys_mount(uint32 cmd, uint8 rdrive)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	pp_switch_to_write();

	return error;
}

uint16 sys_chdir(uint32 cmd, char far* dirname)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, dirname);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	pp_switch_to_write();


	return error;
}

uint16 sys_findfirst(uint32 cmd, char far* filespec, uint16 attr, DIRREC_PTR dirrec, SRCHREC_PTR srchrec)
{
	uint32	size;
	uint32	time;
	uint8	attrib;
	uint16	error;
	uint16	len;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, filespec);
	xdr_uint16(XDR_WRITE, &attr);

	pp_switch_to_read();


	xdr_uint16(XDR_READ, &error);

	if (!error)
	{
		xdr_direntry(XDR_READ, dirrec->file_name, &size, &attrib, (uint16*) &time, ((uint16*) &time) +1);
		xdr_data(XDR_READ, ((char far *) srchrec)+1, &len);
		dirrec->file_size=size;
		dirrec->file_time=time;
		dirrec->file_attr=attrib;
	}

	pp_switch_to_write();

	return error;
}

uint16 sys_findnext(uint32 cmd, SRCHREC_PTR srchrec, DIRREC_PTR dirrec)
{
	uint16	error;
	uint16	len;
	uint8	attrib;
	uint32	size;
	uint32	time;

	xdr_uint32(XDR_WRITE, &cmd);
	len = sizeof(SRCHREC)-1;
	xdr_data(XDR_WRITE, ((char far *) srchrec) + 1, &len);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	if (!error)
	{
		xdr_direntry(XDR_READ, dirrec->file_name, &size, &attrib, (uint16*) &time, ((uint16*) &time) +1);
		xdr_data(XDR_READ, ((char far *) srchrec) + 1, &len);
		dirrec->file_size=size;
		dirrec->file_time=time;
		dirrec->file_attr=attrib;
	}

	pp_switch_to_write();

	return error;
}

uint16 sys_rmdir(uint32 cmd, char far* dirname)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, dirname);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	pp_switch_to_write();


	return error;
}

uint16 sys_mkdir(uint32 cmd, char far* dirname)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, dirname);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	pp_switch_to_write();


	return error;
}

uint16 sys_getdiskspace(uint32 cmd, uint16* sec, uint16* clusters, uint16* bps, uint16* freeclus)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);
	if (!error)
	{
		xdr_uint16(XDR_READ, sec);
		xdr_uint16(XDR_READ, clusters);
		xdr_uint16(XDR_READ, bps);
		xdr_uint16(XDR_READ, freeclus);
	}

	pp_switch_to_write();


	return error;
}

uint16 sys_delfile(uint32 cmd, char far* filename)
{
	char far* pc;
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
		/*
		 * dos changes directory to the right place and issues
		 * a delfile with full path. There is 1) no need to send the path
		 * and 2) the system call we use on the server side won't accept
		 * the path name
		 */
	if (pc=_fstrrchr(filename, '\\'))
		xdr_filename(XDR_WRITE, pc+1);
	else
		xdr_filename(XDR_WRITE, filename);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	pp_switch_to_write();


	return error;
}

uint16 sys_renfile(uint32 cmd, char far* curfile, char far* newfile)
{
	/* not supported yet */
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, curfile);
	xdr_filename(XDR_WRITE, newfile);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	pp_switch_to_write();


	return error;
}

uint16 sys_setfileattr(uint32 cmd, char far* filename, unsigned attr)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, filename);
	xdr_uint16(XDR_WRITE, &attr);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	pp_switch_to_write();


	return error;
}

uint16 sys_getfileattr(uint32 cmd, char far* filename, unsigned *attr)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, filename);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);
	if (!error)
		xdr_uint16(XDR_READ, attr);

	pp_switch_to_write();


	return error;
}

uint16 sys_openfile(uint32 cmd, char far *filename, unsigned mode, unsigned* handle)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, filename);
	xdr_uint16(XDR_WRITE, &mode);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);
	if (!error)
		xdr_uint16(XDR_READ, handle);

	pp_switch_to_write();


	return error;
}

uint16 sys_special_openfile(uint32 cmd, char far *filename, unsigned attrib, unsigned open_mode, unsigned action, unsigned* result, unsigned* handle)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, filename);
	xdr_uint16(XDR_WRITE, &attrib);
	xdr_uint16(XDR_WRITE, &open_mode);
	xdr_uint16(XDR_WRITE, &action);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);
	if (!error)
	{
		xdr_uint16(XDR_READ, result);
		xdr_uint16(XDR_READ, handle);
	}


	pp_switch_to_write();


	return error;
}

uint16 sys_createfile(uint32 cmd, char far *filename, unsigned attr, unsigned* handle)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_filename(XDR_WRITE, filename);
	xdr_uint16(XDR_WRITE, &attr);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);
	if (!error)
		xdr_uint16(XDR_READ, handle);


	pp_switch_to_write();

	return error;
}


uint16 sys_closefile(uint32 cmd, unsigned handle)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_uint16(XDR_WRITE, &handle);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	pp_switch_to_write();


	return error;
}

uint16 sys_commitfile(uint32 cmd, unsigned handle)
{
	uint16	error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_uint16(XDR_WRITE, &handle);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	pp_switch_to_write();


	return error;
}
static uint16 do_readfile(uint32 cmd, unsigned handle, uint32 file_pos, unsigned size)
{
	uint16		error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_uint16(XDR_WRITE, &handle);
	xdr_uint32(XDR_WRITE, &file_pos);
	xdr_uint16(XDR_WRITE, &size);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);

	return error;
}

/*
 * Steps for reading, we set a minimum and maximum on the size of the
 * buffer we request for reading as the server needs some space to
 * compress
 * 1 determine how much we read (input_size)
 * 2 make the read request at the server and read an errno, if errno set
 *		we're done
 * 3 depending on the whether or not the data is compressed we read in 
 * 		the right buffer
 * 4 if the data is compressed, uncompress it
 * 5 if there is more to read, do so
 */
uint16 sys_readfile(uint32 cmd, unsigned handle, uint32 file_pos, unsigned size, char far *buffer, unsigned *nbytes)
{
	uint16		nbytes2;
	uint16		input_size;
	uint32		uncompress_len;
	uint16		error;
	uint8		iscompressed;

	MAPBUFFER
	//lzrw1_sethash(emsPageFrame+ 0xFFFFu - LZRW1_HASHSIZE*sizeof(uint8 far *));

	*nbytes=0;
	do
	{
			/*
			 * Step 1: determine how much we request for reading
			 */
		if (!use_compress || size < MIN_COMPRESS)
		{ 	
			cmd = cmd & 0x0000FFFFul;
			input_size = size;
			iscompressed = 0;
		}
		else
		{
			cmd = (cmd & 0x0000FFFFul) | 0x00010000ul;
			input_size = (size < LZRW1_INBLKSIZE) ? size : LZRW1_INBLKSIZE;
			iscompressed = 1;
		}

			/*
			 * Step 2: make the command at the server and see what (s)he
			 *	responds
			 */
		if ((error=do_readfile(cmd, handle, file_pos, input_size)))
		{
			pp_switch_to_write();
			return error;
		}

			/*
			 * Step 3: see where to read in the data
			 */
		if (iscompressed)
		{
			xdr_data(XDR_READ, emsPageFrame, &nbytes2);

				/*
				 * Step 4: uncompress the received buffer
				 */
			lzrw1_decompress(emsPageFrame, nbytes2, buffer, &uncompress_len);
			nbytes2= (unsigned) uncompress_len;
		}
		else
			xdr_data(XDR_READ, buffer, &nbytes2);

		pp_switch_to_write();
		
			/*
			 * Step 5: see if there is more to read
			 */
		*nbytes+=nbytes2;			// adjust the bytes we read succesfully
		file_pos+=nbytes2;			// adjust filepointer
		buffer+=nbytes2;			// adjust the input buffer pointer
		size-=nbytes2;				// adjust the number of bytes still to go
	} while (size && nbytes2);		

	return error;
}

static uint16 do_writefile(uint32 cmd, unsigned handle, uint32 file_pos, uint32 time, unsigned size, char far* buffer, unsigned *nbytes)
{
	uint16		error;

	xdr_uint32(XDR_WRITE, &cmd);
	xdr_uint16(XDR_WRITE, &handle);
	xdr_uint32(XDR_WRITE, &file_pos);
	xdr_uint32(XDR_WRITE, &time);
	xdr_data(XDR_WRITE, buffer, &size);

	pp_switch_to_read();

	xdr_uint16(XDR_READ, &error);
	if (!error)
		xdr_uint16(XDR_READ, nbytes);

	pp_switch_to_write();

	return error;
}

/*
 * Steps in writing to file. Because of the limits on the compression buffers
 * we set a maximum and minimum size on the buffer we compress
 *
 * 1 determine how much we are sending
 * 2 if we should compress, do so
 * 3 send the data to the server
 * 4 if there is more to write, do so
 */
uint16 sys_writefile(uint32 cmd, unsigned handle, uint32 file_pos, uint32 time, unsigned size, char far* buffer, unsigned *nbytes)
{
	char far 	*send_buffer;		// pointer to what we are sending
	uint16		nbytes2;			// what we the server wrote
	uint16		input_size;			// what we like to send
	uint16		send_size;			// size of what we are sending
	uint32		compress_len;		// compressed length
	uint16		error;			
		
	MAPBUFFER
	lzrw1_sethash((unsigned long) emsPageFrame+ 0xFFFFu - LZRW1_HASHSIZE*sizeof(uint8 far *));

		/*
		 * If we don't clear the hash table, the algorithm woun't
		 * work properly :-(
		 */
	_fmemset(emsPageFrame+ 0xFFFFu - LZRW1_HASHSIZE*sizeof(uint8 far *), 0, LZRW1_HASHSIZE*sizeof(uint8 far *));

	*nbytes=0;
	do
	{
			/*
			 * Step 1 determine how much we write and from where
			 */
		if (!use_compress || size < MIN_COMPRESS)
		{
			cmd = cmd & 0x0000FFFFul;
			input_size = size;
			send_size=input_size;
			send_buffer=buffer;
		}
		else
		{

				/*
				 * Step 2 if we should compress, do it
				 */
			cmd = (cmd & 0x0000FFFFul) | 0x00010000ul;
			input_size = (size < LZRW1_INBLKSIZE) ? size : LZRW1_INBLKSIZE;
			lzrw1_compress(buffer, input_size, emsPageFrame, &compress_len);
			send_buffer=emsPageFrame;
			send_size= (unsigned) compress_len;
		}

			/*
			 * Step 3 Send the buffer (compressed or not)
			 */
		if ((error=do_writefile(cmd, handle, file_pos, time, send_size, send_buffer, &nbytes2)))
			return error;

			/*
			 * Step 4 see if there is more to write
			 */
		*nbytes+=nbytes2;			// adjust the bytes we wrote succesfully
		file_pos+=nbytes2;			// adjust filepointer
		buffer+=nbytes2;			// adjust the input buffer pointer
		size-=nbytes2;				// adjust the number of bytes still to go
	} while (size && nbytes);

	return error;
}
