#ifndef __REDIR_H
#define __REDIR_H

#include "systypes.h"

typedef void (far *FARPROC)(void);
typedef void (*PROC)(void);

#pragma pack(1)

/* FindFirst/Next data block - ALL DOS VERSIONS */
typedef struct
{
        uint8           drive_no;
        char            srch_mask[11];
        uint8           attr_mask;
        uint16          dir_entry_no;
        uint16          dir_sector;
        uint8           f1[4];
} SRCHREC, far* SRCHREC_PTR;

/* DOS System File Table entry - ALL DOS VERSIONS */
// Some of the fields below are defined by the redirector, and differ
// from the SFT normally found under DOS
typedef struct
{
        uint16          handle_count;
        uint16          open_mode;
        uint8           file_attr;
        uint16          dev_info_word;
        uint8 far *     dev_drvr_ptr;
        uint16          start_sector;
        uint32          file_time;
        int32           file_size;
        int32           file_pos;
        uint16          rel_sector;
        uint16          abs_sector;
        uint16          dir_sector;
        uint8           dir_entry_no;
        char            file_name[11];
} SFTREC, far* SFTREC_PTR;

/* DOS Current directory structure - DOS VERSION 3.xx */
typedef struct
{
        char            current_path[67];
        uint16          flags;
        uint8           f1[10];
        uint16          root_ofs;
} V3_CDS, far* V3_CDS_PTR;


/* DOS Current directory structure - DOS VERSION 4.xx */
typedef struct
{
        char            current_path[67];
        uint16          flags;
        uint8           f1[10];
        uint16          root_ofs;
        uint8           f2[7];
} V4_CDS, far* V4_CDS_PTR;


/* DOS Directory entry for 'found' file - ALL DOS VERSIONS */
typedef struct
{
        char            file_name[11];
        uint8           file_attr;
        uint8           f1[10];
        uint32          file_time;
        uint16          start_sector;
        int32           file_size;
} DIRREC, far* DIRREC_PTR;

/* Swappable DOS Area - DOS VERSION 3.xx */
typedef struct
{
        uint8           f0[12];
        uint8 far *     current_dta;
        uint8           f1[30];
        uint8           dd;
        uint8           mm;
        uint16          yy_1980;
        uint8           f2[96];
        char            file_name[128];
        char            file_name_2[128];
        SRCHREC         srchrec;
        DIRREC          dirrec;
        uint8           f3[81];
        char            fcb_name[11];
        uint8           f4;
        char            fcb_name_2[11];
        uint8           f5[11];
        uint8           srch_attr;
        uint8           open_mode;
        uint8           f6[48];
        uint8 far *     cdsptr;
        uint8           f7[72];
        SRCHREC         rename_srchrec;
        DIRREC          rename_dirrec;
} V3_SDA, far* V3_SDA_PTR;


/* Swappable DOS Area - DOS VERSION 4.xx */
typedef struct
{
        uint8           f0[12];
        uint8 far *     current_dta;
        uint8           f1[32];
        uint8           dd;
        uint8           mm;
        uint16          yy_1980;
        uint8           f2[106];
        char            file_name[128];
        char            file_name_2[128];
        SRCHREC         srchrec;
        DIRREC          dirrec;
        uint8           f3[88];
        char            fcb_name[11];
        uint8           f4;
        char            fcb_name_2[11];
        uint8           f5[11];
        uint8           srch_attr;
        uint8           open_mode;
        uint8           f6[51];
        uint8 far *     cdsptr;
        uint8           f7[87];
        uint16          action_2E;
        uint16          attr_2E;
        uint16          mode_2E;
        uint8           f8[29];
        SRCHREC         rename_srchrec;
        DIRREC          rename_dirrec;
} V4_SDA, far* V4_SDA_PTR;


/* DOS List of lists structure - DOS VERSIONS 3.1 thru 4 */
/* We don't need much of it. */
typedef struct
{
        uint8           f1[22];
        V3_CDS_PTR      cds_ptr;
        uint8           f2[7];
        uint8           last_drive;
} LOLREC, far* LOLREC_PTR;


/* DOS 4.00 and above lock/unlock region structure */
/* see lockfil() below (Thanks to Martin Westermeier.) */
typedef struct
{
        uint32          region_offset;
        uint32          region_length;
        uint8           f0[13];
        char            file_name[80]; // 80 is a guess
} LOCKREC, far* LOCKREC_PTR;


/* The following structure is compiler specific, and maps
        onto the registers pushed onto the stack for an interrupt
        function. */
typedef struct
{
#ifdef __BORLANDC__
        uint16          bp, di, si, ds, es, dx, cx, bx, ax;
#else
        uint16          es, ds, di, si, bp, sp, bx, dx, cx, ax;
#endif
        uint16          ip,cs,flags;
} ALL_REGS;

#pragma pack()

/* all the calls we need to support are in the range 0..2Eh */
/* This serves as a list of the function types that we support */
#define         _inquiry                0x00
#define         _rd                     0x01
#define         _md                     0x03
#define         _shutdown               0x04
#define         _cd                     0x05
#define         _closefile              0x06
#define         _commitfile             0x07
#define         _readfile               0x08
#define         _writefile              0x09
#define         _lockfile               0x0A
#define         _unlockfile             0x0B
#define         _diskspace              0x0C
#define         _setfileattr            0x0E
#define         _getfileattr            0x0F
#define         _renamefile             0x11
#define         _deletefile             0x13
#define         _openfile               0x16
#define         _createfile             0x17
#define         _findfirst              0x1B
#define         _findnext               0x1C
#define         _seekfromend            0x21
#define         _unknown_fxn_2D 		0x2D
#define         _specialopenfile        0x2E
#define         _unsupported    		0xFF




#define ATTR_READONLY	0x01
#define ATTR_HIDDEN		0x02
#define ATTR_SYSTEM		0x04
#define ATTR_VOLUME		0x08
#define ATTR_SUBDIR		0x10
#define ATTR_ARCHIVE	0x20





#endif
