#pragma check_stack(off)
#include <dos.h>
#include <conio.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <direct.h>

#include "systypes.h"
#include "vsprintf.h"
#include "tsr.h"
#include "ems.h"
#include "debug.h"
#include "xdr.h"
#include "pport.h"
#include "lzrw1.h"


typedef void (far *FARPROC)(void);
typedef void (*PROC)(void);

#ifndef MK_FP
#define MK_FP(a,b)  ((void far *)(((unsigned long)(a) << 16) | (b)))
#endif

char *usage_string = "Usage:   rmount localdisk=remotedisk etc.\r\n\r\n"
	"Options:\r\n"
	"   DC  : Debug to console (this is the default)\r\n"
	"   DEx : Debug to ems, use x pages (defaults to 1)\r\n"
	"   Lx  : Debug level x (defaults to 0)\r\n"
	"   I   : Interactive mode\r\n"
	"   SF  : Show list of debug flags\r\n"
	"   Px  : Use LPTx (defaults to 0)\r\n"
	"\r\n";


char *TsrName="srvmount";
char *VersionString="1.0";

/* ****************************************************
   Constants and Macros
   **************************************************** */

#define		TRUE				1
#define		FALSE				0

#define		MAX_FXN_NO			0x2E

	/*
	 * Mounttable, we only allow a drive to be mounted once,
	 * there may be multiple drive mounted at the same time however
	 */
static char			MountTable[26];

	/*
	 * remember our starting place, so we can go back
	 * when we're done.
	 */
static char			old_dir[128];
static int			old_drive;

	/*
	 * Some neat statitstics, use /l1 on commandline (debug info SYSTEM)
	 * to see some compression ratio's
	 */
static uint32		nRead;
static uint32		nRCompressed;
static uint32		nRNotCompressed;
static uint32		nWrote;
static uint32		nWCompressed;
static uint32		nWNotCompressed;

	/*
	 * within handling requests we need to know which command on which
	 * drive we're handling.
	 */
static uint32		cur_cmd;
static uint8		cur_drive;
static int			ShutdownRequest=0;


	/*
	 * Some globals for temporary buffers
	 */
static unsigned		seg_rwbuffer;	// segment of our 64 KB read/write buffer
static char far*	rwBuffer;		// far pointer to our read/write buffer
static unsigned		seg_compbuffer;	// segment of our 64 KB compression buffer
static char far*	compBuffer;		// far pointer to our compression buffer
static char			buf1[128];
static char			buf2[128];

	/*
	 * used for find first/find next calls
static	struct _find_t	findlist[26];
	 */
static	char			substList[26];
static	struct _find_t	findtemp;
static union REGS		serverregs;

void cmd_inquiry(void)				
{
	int			dout;
	uint16		error=0;

	pp_switch_to_write();
	if (MountTable[cur_drive])
	{
		error=1;
		xdr_uint16(XDR_WRITE, &error);
		dprintf(D_ALWAYS, "Mounting %c: failed, drive already mounted\n", cur_drive+'A');
		return;
	}

	error=_chdir("\\");
	xdr_uint16(XDR_WRITE, &error);

	dout=debug_setout(DOUT_CONSOLE);
	dprintf(D_ALWAYS, "Mounting %c: %s\n", cur_drive+'A', error? "failed":"succeeded");
	debug_setout(dout);

	if (!error)
		MountTable[cur_drive]=1;
}

/*
 * convert a zero-terminated filenamestring into a Dos fcb filename
 * (11 chars) without a zero terminating character and the extention dot.
 */

static void mk_fcb_filename(char *filename, char *buf)
{
	int		i=0;
	char	c;

		/*
		 * Copy filename part
		 */
	while ( (c=*filename++) ) 
	{
			/*
			 * if it's either '.' or '..' we copy them otherwise it's 
			 * the dot to denote the extention part, so break
			 */
		if (c=='.')
		{
			if (i>1)
				break;
			else
				if (i==1 && *(filename-2)!='.')
					break;
		}	
		i++;
		*buf++=c;
	}

		/*
		 * pad with spaces
		 */
	while (i < 8)
	{
		i++;
		*buf++=' ';
	}
		/*
		 * copy extention part
		 */
	if (c=='.')
	{
		while ( (c=*filename++) )
		{
			i++;
			*buf++=c;
		}
	}

		/*
		 * pad with spaces
		 */
	while (i++ < 11)
		*buf++=' ';
}

void cmd_findnext(void)				
{
	uint16				error;
	uint16				len;
	struct _find_t*		findentry=&findtemp;

	dprintf(D_SERVER, "%c: finding next...\n", cur_drive+'A');

	xdr_data(XDR_READ, (char far *) &findtemp.reserved[1],  &len);
	//findtemp.reserved[0]= cur_drive + 1;
	findtemp.reserved[0]=substList[cur_drive];
	error=_dos_findnext(findentry);

	pp_switch_to_write();

	xdr_uint16(XDR_WRITE, &error);
	if (!error)
	{
		mk_fcb_filename(findentry->name, buf1);
		xdr_direntry(XDR_WRITE, buf1, &findentry->size, &findentry->attrib, &findentry->wr_time, &findentry->wr_date);
		len = sizeof(findtemp.reserved)-1;
		xdr_data(XDR_WRITE, (char far *) &findtemp.reserved[1], &len);
        dprintf(D_SERVER, "findnext succeeded, entry=%11s, attr=0x%x, size=%lu\n", buf1, (unsigned) findentry->attrib, findentry->size);
	}
	else
		dprintf(D_SERVER, "finding next failed, error=0x%x\n", error);
}

void cmd_findfirst(void)
{
	uint16				mask;
	uint16				len;
	uint16				error;
	struct _find_t*		findentry;

	findentry= &findtemp;

	xdr_filename(XDR_READ, buf1);
	xdr_uint16(XDR_READ, &mask);

    dprintf(D_SERVER,"%c: findfirst:%s, mask=0x%x\n", cur_drive+'A', buf1, mask);

	error=_dos_findfirst(buf1, mask, findentry);
	substList[cur_drive]=findtemp.reserved[0];

	pp_switch_to_write();

	xdr_uint16(XDR_WRITE, &error);
	if (!error)
	{
		mk_fcb_filename(findentry->name, buf1);
		xdr_direntry(XDR_WRITE, buf1, &findentry->size, &findentry->attrib, &findentry->wr_time, &findentry->wr_date);
		len = sizeof(findtemp.reserved)-1;
		xdr_data(XDR_WRITE, (char far *) &findtemp.reserved[1], &len);
        dprintf(D_SERVER, "findfirst succeeded, entry=%11s, attr=0x%x, size=%lu\n", buf1, (unsigned) findentry->attrib, findentry->size);
	}
	else
    	dprintf(D_SERVER, "findfirst failed, error=0x%x\n", error);
}


void cmd_rmdir(void)
{
	uint16		error;

	xdr_filename(XDR_READ, buf1);
	dprintf(D_SERVER, "%c: rmdir:%s\n", cur_drive+'A', buf1);

	pp_switch_to_write();


	if (!(error=_rmdir(buf1)))
	{
		xdr_uint16(XDR_WRITE, &error);
        dprintf(D_SERVER,"rmdir succeeded\n");
	}
	else
	{
		xdr_uint16(XDR_WRITE, &((uint16) _doserrno));
        dprintf(D_SERVER,"rmdir failed, error=0x%x\n", (uint16) _doserrno);
	}
}

void cmd_mkdir(void)
{
	uint16		error;

	xdr_filename(XDR_READ, buf1);
	dprintf(D_SERVER, "%c: mkdir:%s\n", cur_drive+'A', buf1);


	pp_switch_to_write();

	if (!(error=_mkdir(buf1)))
	{
		xdr_uint16(XDR_WRITE, &error);
        dprintf(D_SERVER,"mkdir succeeded\n");
	}
	else
	{
		xdr_uint16(XDR_WRITE, &((uint16) _doserrno));
		dprintf(D_SERVER,"mkdir failed, error=0x%x\n", (uint16) _doserrno);
	}
}

void cmd_commitfile(void)
{
	uint16		error, handle;

	xdr_uint16(XDR_READ, &handle);
	dprintf(D_SERVER, "%c: commitfile, handle=0x%x\n", cur_drive+'A', handle);

	pp_switch_to_write();


	if (!(error=_dos_commit(handle)))
	{
		xdr_uint16(XDR_WRITE, &error);
        dprintf(D_SERVER,"commitfile succeeded\n");
	}
	else
	{
		xdr_uint16(XDR_WRITE, &((uint16) _doserrno));
		dprintf(D_SERVER,"commitfile failed, error=0x%x\n", (uint16) _doserrno);
	}
}


void cmd_cd(void)
{
	uint16		error;

	xdr_filename(XDR_READ, buf1);

		/*
		 * Now, this is funny. The normal _chdir returns only failed or 
		 * succeeded without specifying why it failed (just ENOENT). However 
		 * if we delete a file with Dos's 'del filename', Dos first gonna 
		 * check if it's a directory or not; it does so by trying to chdir 
		 * to it! Since the servers _chdir always returns a ENOENT on a 
		 * non-directory, Dos not only assumes that the 'filename'
		 * is not a directory but that 'filename' does not exists at all.
		 * That's why we issue the system call ourselves here, to obtain
		 * the correct failure code.
		 */
	dprintf(D_SERVER, "%c: chdir:%s\n", cur_drive+'A', buf1);

	serverregs.h.ah=0x3B;
	serverregs.x.dx=(uint16) &buf1;
	int86(0x21, &serverregs, &serverregs);

	pp_switch_to_write();


	if (!(error=serverregs.x.cflag))
	{
		xdr_uint16(XDR_WRITE, &error);
        dprintf(D_SERVER, "chdir succeeded\n");
	}
	else
	{
		xdr_uint16(XDR_WRITE, &serverregs.x.ax);
		dprintf(D_SERVER, "chdir failed, error=0x%x\n", error);
	}
}

void cmd_setfileattr(void)
{
	uint16		error;
	uint16		attr;

	xdr_filename(XDR_READ, buf1);
	xdr_uint16(XDR_READ, &attr);

	pp_switch_to_write();


	dprintf(D_SERVER, "%c: setfileattr:%s, attr=0x%x\n", cur_drive+'A', buf1, attr);

	error=_dos_setfileattr(buf1, attr);
	if (error)
		dprintf(D_SERVER, "setfileattr failed, error=0x%x\n", error);
	else
        dprintf(D_SERVER, "setfileattr succeeded\n");

	xdr_uint16(XDR_WRITE, &error);
}


void cmd_renfile(void)
{
/*
	not supported as we do not have Dos 6.x with it's move command
	I cannot test it, sorry.
	uint16		error;

	xdr_filename(XDR_READ, buf1);
	xdr_filename(XDR_READ, buf2);

	pp_switch_to_write();

	if (!(error=rename(buf1, buf2)))
		xdr_uint16(XDR_WRITE, &error);
	else
		xdr_uint16(XDR_WRITE, &((uint16) _doserrno));
*/
}

void cmd_delfile(void)
{
		/*
		 * Dos depends on the older fcb inetrface to delete files.
		 * So, we make up a fcb; clumsy I'll admit but allas.
		 */
	xdr_filename(XDR_READ, buf1);
	mk_fcb_filename(buf1, buf2+1);	// mae it an fcb filename
	buf2[0];		// first byte is the drive number, zero is current
	memset(buf2+12, 0, 25);

	dprintf(D_SERVER, "%c: delfile:%s\n", cur_drive+'A', buf1);

	serverregs.h.ah=0x13;
	serverregs.x.dx=(uint16) &buf2;
	int86(0x21, &serverregs, &serverregs);
	serverregs.h.ah=0;

	pp_switch_to_write();


	xdr_uint16(XDR_WRITE, &serverregs.x.ax);
	if (serverregs.x.ax)
		dprintf(D_SERVER,"delfile failed, error=0x%x\n", serverregs.x.ax);
	else
        dprintf(D_SERVER,"delfile succeeded\n");
}

void cmd_createfile(void)
{
	uint16		error;
	uint16		handle;
	uint16		attr;

	xdr_filename(XDR_READ, buf1);
	xdr_uint16(XDR_READ, &attr);

	dprintf(D_SERVER, "%c: createfile:%s, attr=0x%x\n", cur_drive+'A', buf1, attr);

	pp_switch_to_write();


	if (!(error=_dos_creat(buf1, attr, &handle)))
	{
		xdr_uint16(XDR_WRITE, &error);
		xdr_uint16(XDR_WRITE, &handle);
        dprintf(D_SERVER,"createfile succeeded, handle=0x%x \n", handle);
	}
	else
	{
		xdr_uint16(XDR_WRITE, &((uint16) _doserrno));
		dprintf(D_SERVER,"createfile failed, error=0x%x\n", (uint16) _doserrno);
	}
}


void cmd_closefile(void)
{
	uint16		error, handle;

	xdr_uint16(XDR_READ, &handle);

	dprintf(D_SERVER, "%c: closefile, handle=0x%x\n", cur_drive+'A', handle); 

	pp_switch_to_write();


	if (!(error=_dos_close(handle)))
	{
		xdr_uint16(XDR_WRITE, &error);
        dprintf(D_SERVER,"closefile succeeded\n");
	}
	else
	{
		xdr_uint16(XDR_WRITE, &((uint16) _doserrno));
		dprintf(D_SERVER,"closefile failed, error=0x%x\n", (uint16) _doserrno);
	}
}

static uint16 set_filepointer(unsigned handle, uint32 file_pos)
{
	serverregs.h.ah=0x42;
	serverregs.h.al=0x00;
	serverregs.x.bx=handle;
	serverregs.x.cx=*(((unsigned *) &file_pos) + 1);
	serverregs.x.dx=*((unsigned *) &file_pos);
	int86(0x21, &serverregs, &serverregs);
	return serverregs.x.cflag ? serverregs.x.ax : 0;
}

/*
 * Steps in writing to file
 * 1 get handle to write to, position to write to, time stamp and number of 
 *		bytes to write
 * 2 in the second leftmost byte of 'cmd' we get a flag wether the buffer 
 *		is compressed or not
 * 3 read the data we should write to file (which could be compressed)
 * 4 if the data is compressed, uncompress it
 * 5 position the file pointer, if error send the dos errno and we're done
 * 6 write the data to file , if error, send back dos errno and we're done
 * 7 send back the number of bytes we wrote
 */ 
void cmd_writefile(void)
{
	uint16		error, handle, size, count;
	uint32		file_pos, time;
	uint8		compressed;
	uint32		compress_len;
	char far *	write_buffer;
	uint16		write_size;

	dprintf(D_SERVER, "%c: writefile, receiving data...\r", cur_drive+'A');

		/*
		 * Step 1: read parameters
		 */
	xdr_uint16(XDR_READ, &handle);
	xdr_uint32(XDR_READ, &file_pos);
	xdr_uint32(XDR_READ, &time);
	xdr_data(XDR_READ, rwBuffer, &size);

	pp_switch_to_write();


		/*
		 * Step 2: see if client sends it compressed
		 */
	compressed= (uint8) ((cur_cmd & 0x00FF0000ul) >> 16);

	dprintf(D_SERVER, "%c: writefile%s, handle=0x%x, file_pos=%lu, count=%u\n", cur_drive+'A', compressed?" (compressed)":"", handle, file_pos, size);

		/*
		 * Step 3: if it's compressed, uncompress it
		 */
	if (compressed)
	{
		nRCompressed += size;
		lzrw1_decompress(rwBuffer, size, compBuffer, &compress_len);
		write_buffer = compBuffer;
		write_size = (unsigned) compress_len;
	}
	else
	{
		nRNotCompressed += size;
		write_buffer = rwBuffer;
		write_size = size;
	}
	nRead += write_size;

		/*
		 * Step 5: position fileponter
		 */
	if ((error = set_filepointer(handle, file_pos)))
	{
		xdr_uint16(XDR_WRITE, &error);
		dprintf(D_SERVER, "writefile failed, could not move filepointer\n");
		return;
	}

		/*
		 * Step 6: write to file
		 */
	if (!(error = _dos_write(handle, write_buffer, write_size, &count)))
	{
			/*
			 * Step 7: tell client how much we wrote
			 */
		xdr_uint16(XDR_WRITE, &error);
		xdr_uint16(XDR_WRITE, &count);
        dprintf(D_SERVER,"writefile succeeded, wrote %u bytes\n", count);
	}
	else
	{
		xdr_uint16(XDR_WRITE, &((uint16) _doserrno));
		dprintf(D_SERVER, "writefile failed, error=0x%x\n", (uint16) _doserrno);
	}
}

/*
 * Steps to read from file:
 * 1 get handle to read from, position to read from and number of bytes to read
 * 2 in the second leftmost byte of 'cmd' we get a flag whether to send a
 *		compressed buffer or not
 * 3 position the file pointer, if error send the errno and we're done
 * 4 read the data from file, if error, send errno and we're done
 * 5 if we should compress, do so 
 * 6 send the data (compressed or not)
 */
void cmd_readfile(void)
{
	uint16		error, handle, size, count;
	uint32		file_pos;
	uint8		compressed;
	char far *	send_buffer;
	uint32		compress_len;
	uint16		send_size;

		/*
		 * Step 1: read parameters
		 */
	xdr_uint16(XDR_READ, &handle);
	xdr_uint32(XDR_READ, &file_pos);
	xdr_uint16(XDR_READ, &size);

	pp_switch_to_write();


		/*
		 * Step 2: see if client wants it compressed
		 */
	compressed = (uint8) ((cur_cmd & 0x00FF0000ul) >> 16);

	dprintf(D_SERVER, "%c: readfile%s, handle=0x%x, file_pos=%lu, count=%u\n", cur_drive+'A', compressed?" (compressed)":"", handle, file_pos, size);

		/*
		 * Step 3: Position file pointer
		 */
	if ((error=set_filepointer(handle, file_pos)))
	{
		xdr_uint16(XDR_WRITE, &error);
		dprintf(D_SERVER, "readfile failed, could not move filepointer\n");
		return;
	}

		/*
		 * Step 4: read from file
		 */
	if ((error=_dos_read(handle, rwBuffer, size, &count)))
	{
		xdr_uint16(XDR_WRITE, &((uint16) _doserrno));
		dprintf(D_SERVER,"readfile failed, error=0x%x\n", (uint16) _doserrno);
		return;
	}

	nWrote += count;
	xdr_uint16(XDR_WRITE, &error);

		/*
		 * Step 5: compress the rwBuffer into the compBuffer
		 */
	if (compressed)
	{
		lzrw1_compress(rwBuffer, count, compBuffer, &compress_len);
		send_buffer = compBuffer;
		send_size = (unsigned) compress_len;
		nWCompressed += send_size;
	}
	else
	{
		send_buffer=rwBuffer;
		send_size=count;
		nWNotCompressed += send_size;
	}

		/*
		 * Step 6: send the buffer, compressed or not
		 */
	dprintf(D_SERVER, "readfile succeeded, writing %u bytes...\r", send_size);
	xdr_data(XDR_WRITE, send_buffer, &send_size);
	dprintf(D_SERVER, "readfile succeeded, writing %u bytes, done\n", send_size);
}

void cmd_diskspace(void)
{
	uint16				error;
	struct _diskfree_t	diskfree;

	dprintf(D_SERVER,"%c: diskspace\n", cur_drive+'A');

	pp_switch_to_write();


	error=_dos_getdiskfree(cur_drive+1, &diskfree);
	xdr_uint16(XDR_WRITE, &error);
	if(!error)
	{
		xdr_uint16(XDR_WRITE, &diskfree.sectors_per_cluster);
		xdr_uint16(XDR_WRITE, &diskfree.total_clusters);
		xdr_uint16(XDR_WRITE, &diskfree.bytes_per_sector);
		xdr_uint16(XDR_WRITE, &diskfree.avail_clusters);
        dprintf(D_SERVER,"diskspace succeeded\n");
	}
	else
		dprintf(D_SERVER,"diskspace failed, error=0x%x\n", error);

}

void cmd_getfileattr(void)
{
	uint16		error;
	uint16		attr;

	xdr_filename(XDR_READ, buf1);

	dprintf(D_SERVER,"%c: getfileattr:%s\n", cur_drive+'A', buf1);

	pp_switch_to_write();


	error=_dos_getfileattr(buf1, &attr);
	xdr_uint16(XDR_WRITE, &error);
	if (!error)
	{
		xdr_uint16(XDR_WRITE, &attr);
        dprintf(D_SERVER, "getfileattr succeeded, attr=0x%x\n", attr);
	}
	else
		dprintf(D_SERVER, "getfileattr failed, error=0x%x\n", error);
}

void cmd_openfile(void)
{
	uint16		error;
	uint16		handle;
	uint16		mode;

	xdr_filename(XDR_READ, buf1);
	xdr_uint16(XDR_READ, &mode);

	dprintf(D_SERVER, "%c: openfile:%s, mode=0x%x\n", cur_drive+'A', buf1, mode);

	pp_switch_to_write();


	if (!(error=_dos_open(buf1, mode, &handle)))
	{
		xdr_uint16(XDR_WRITE, &error);
		xdr_uint16(XDR_WRITE, &handle);
        dprintf(D_SERVER, "openfile succeeded, handle=0x%x\n", handle);
	}
	else
	{
		xdr_uint16(XDR_WRITE, &((uint16) _doserrno));
		dprintf(D_SERVER,"openfile failed, error=0x%x\n", (uint16) _doserrno);
	}
}

void cmd_special_openfile(void)
{
	uint16		error;

	xdr_filename(XDR_READ, buf1);
	xdr_uint16(XDR_READ, &serverregs.x.cx);	// attribute
	xdr_uint16(XDR_READ, &serverregs.x.bx);	// open mode
	xdr_uint16(XDR_READ, &serverregs.x.dx);	// action flag

	pp_switch_to_write();


	dprintf(D_SERVER,"%c: special openfile:%s, mode=0x%x, action=0x%x, attr=0x%x\n", cur_drive+'A', buf1, serverregs.x.bx, serverregs.x.dx, serverregs.x.cx);

	serverregs.h.ah=0x6C;
	serverregs.h.al=0x00;
	serverregs.x.si=(uint16) &buf1;
	int86(0x21, &serverregs, &serverregs);

	if (!(error=serverregs.x.cflag))
	{
		xdr_uint16(XDR_WRITE, &error);
		xdr_uint16(XDR_WRITE, &serverregs.x.cx);	// action result
		xdr_uint16(XDR_WRITE, &serverregs.x.ax);	// handle
        dprintf(D_SERVER, "specialopenfile succeeded, result=0x%x, handle=0x%x\n", serverregs.x.cx, serverregs.x.ax);
	}
	else
	{
		xdr_uint16(XDR_WRITE, &serverregs.x.ax);
		dprintf(D_SERVER, "specialopenfile failed, error=0x%x\n", error);
	}
}


void cmd_seekfromend(void)
{
}

void cmd_closeall(void)
{
        dprintf(D_SERVER, "%c: closeall\n", cur_drive+'A');
        dprintf(D_SERVER, "Oops closeall not implemented yet\n");
}

void cmd_unknown_fxn_2D()
{
	dprintf(D_SERVER,"%c unknown function_2D\n", cur_drive+'A');
}

void cmd_shutdown(void)
{
	dprintf(D_SERVER, "Received shutdown request\n");
	if (nWrote)
	{
		dprintf(D_SYSTEM, "Bytes send on cable:%lu, compressed:%lu, not compressed:%lu\n", nWrote, nWCompressed, nWNotCompressed);
		dprintf(D_SYSTEM, "Bytes read from cable:%lu, compressed:%lu, not compressed:%lu\n", nRead, nRCompressed, nRNotCompressed);
		nRead++;
		nWrote++;		// make sure we don't divide by zero
		dprintf(D_SYSTEM, "Compression ratio's: write:%lu%%, read:%lu%%\n", 
		(unsigned long) ((nWCompressed*100)/(nWrote-nWNotCompressed)), 
		(unsigned long) ((nRCompressed*100)/(nRead-nRNotCompressed)));
	}
	ShutdownRequest=1;
}

void cmd_unsupported(void)
{
	dprintf(D_SERVER,"%c: unsupported function:0x%x\n", cur_drive+'A', 
		(uint16) (cur_cmd & 0x000000FFul));
}

PROC dispatch_table[]=
{
	cmd_inquiry,				/* 0x00h */
	cmd_rmdir,					/* 0x01h */
	cmd_unsupported,			/* 0x02h */
	cmd_mkdir,					/* 0x03h */
	cmd_shutdown,				/* 0x04h */
	cmd_cd,						/* 0x05h */
	cmd_closefile,				/* 0x06h */
	cmd_commitfile,				/* 0x07h */
	cmd_readfile,				/* 0x08h */
	cmd_writefile,				/* 0x09h */
	cmd_unsupported,			/* 0x0Ah */
	cmd_unsupported,			/* 0x0Bh */
	cmd_diskspace,				/* 0x0Ch */
	cmd_unsupported,			/* 0x0Dh */
	cmd_setfileattr,			/* 0x0Eh */
	cmd_getfileattr,			/* 0x0Fh */
	cmd_unsupported,			/* 0x10h */
	cmd_renfile,				/* 0x11h */
	cmd_unsupported,			/* 0x12h */
	cmd_delfile,				/* 0x13h */
	cmd_unsupported,			/* 0x14h */
	cmd_unsupported,			/* 0x15h */
	cmd_openfile,				/* 0x16h */
	cmd_createfile,				/* 0x17h */
	cmd_unsupported,			/* 0x18h */
	cmd_unsupported,			/* 0x19h */
	cmd_unsupported,			/* 0x1Ah */
	cmd_findfirst,				/* 0x1Bh */
	cmd_findnext,				/* 0x1Ch */
	cmd_closeall,				/* 0x1Dh close all files for a process*/
	cmd_unsupported,			/* 0x1Eh */
	cmd_unsupported,			/* 0x1Fh */
	cmd_unsupported,			/* 0x20h */
	cmd_seekfromend,			/* 0x21h */
	cmd_unsupported,			/* 0x22h */
	cmd_unsupported,			/* 0x23h */
	cmd_unsupported,			/* 0x24h */
	cmd_unsupported,			/* 0x25h */
	cmd_unsupported,			/* 0x26h */
	cmd_unsupported,			/* 0x27h */
	cmd_unsupported,			/* 0x28h */
	cmd_unsupported,			/* 0x29h */
	cmd_unsupported,			/* 0x2Ah */
	cmd_unsupported,			/* 0x2Bh */
	cmd_unsupported,			/* 0x2Ch */
	cmd_unsupported,			/* 0x2Dh */
	cmd_special_openfile		/* 0x0Eh */
	};


void handle_requests(int interactive)
{
	int			dout;
	uint8		cmd;
	char		key;

	while(!ShutdownRequest)
	{
		if (interactive)
		{
			key=press_key("Interactive mode: c:go with debug, g:go without debug, other:step next\r\n");
			dout=debug_setout(DOUT_CONSOLE);
			dprintf(D_ALWAYS, "Waiting for command...\r");

			switch (toupper(key))
			{
				case 'C':
					interactive=0;
					dprintf(D_ALWAYS, "Continuing with current debug\n");
					break;
				case 'G':
					interactive=0;
					debug_setlevel(0);
					dprintf(D_ALWAYS, "Going at full speed    \n");
					break;
			}
			debug_setout(dout);
		}

		xdr_uint32(XDR_READ, &cur_cmd);

		cmd=(uint8) (cur_cmd & 0x000000FFl);

			/*
			 * cmd 4 is shutdown, it has no drive
			 */
		if (cmd!=4)
		{
			cur_drive=(uint8) ((cur_cmd & 0x0000FF00l) >> 8 );
			dprintf(D_SERVER, "Receive request %i on drive %c\n", (int) cmd, (int) cur_drive+'A');

			if (_chdrive(cur_drive+1) == -1)
			{
				dprintf(D_ALWAYS, "Could not chdrive to %c, bailing out\n", cur_drive + 'A'-1);

				break;
			}
		}
		dispatch_table[cmd]();

		pp_switch_to_read();

	}
}




/*
 *
 *
 */
int _cdecl main(uint16 argc , char **argv)
{
	int 	dout, dlevel, dpages, showdlist, pport, interactive;

	memset(MountTable, 0, 26);		// nothing mounted
	seg_rwbuffer=seg_compbuffer=0;
	dlevel=showdlist=pport=interactive=dpages=0;
	dout=DOUT_CONSOLE;
	nWrote = nWCompressed = nWNotCompressed = 0;
	nRead = nRCompressed = nRNotCompressed = 0;

		/*
		 * Interpret the arguments
		 */
	for (argv++; *argv; argv++)
	{
		switch (**argv)
		{
			case '-':
			case '/':
				(*argv)++;
				switch (toupper(**argv)) 
				{
					case '?':
						tsr_print_string(usage_string, TRUE);
						return 0;
						break;
					case 'I':
						interactive=1;
						break;
					case 'L':
						dlevel=simple_atoi((*argv)+1);
						break;
					case 'P':
						pport=simple_atoi((*argv)+1);
						break;
					case 'D':
						(*argv)++;
						switch(toupper(**argv))
						{
							case 'C':
								dout=DOUT_CONSOLE;
								break;
							case 'E':
								dout=DOUT_EMS;
								dpages=simple_atoi((*argv)+1);
								if (!dpages)
									dpages++;		// at least 1 page
								break;
							default:
								dout=DOUT_CONSOLE;
								dlevel=0;
								break;
						}
						break;
					case 'S':
						(*argv)++;
						switch(toupper(**argv))
						{
							case 'F':
								showdlist=1;
								break;
							default:
								tsr_print_string("Unrecognized parameter.", TRUE);
								return 0;
						}
						break;
					default:
						tsr_print_string("Unrecognized parameter.", TRUE);
						return 0;
				}
				break;
			default:
				break;
		}
	}

		/*
		 * Show list of debug flags
		 */
	if (showdlist)
	{
		debug_init(0,DOUT_CONSOLE,0);
		debug_showlist();
		dprintf(D_ALWAYS,"\nUse binary | to combine debuglevels\n");
		return 0;
	}

		/*
		 * alloc ems memory
		 */
	if (dout==DOUT_EMS && !emsPresent())
	{
		tsr_print_string("Ems not present, bailing out",TRUE);
		return 0;
	}

		/*
		 * Initialise debugging, but start logging
		 * to console to see the errors
		 */
	debug_init(dlevel, dout, dpages);
	dout=debug_setout(DOUT_CONSOLE);


		/*
		 * init the parallel communication module, and see if
		 * there is anything at the requested LPT port
		 */
	if (!pp_initialize(pport))
	{
		dprintf(D_ALWAYS, "Could not init pp, bailing out.\n");
		return 0;
	}

		/*
		 * remember current drive and path
		 */
	old_drive=_getdrive();
	_getcwd(old_dir, 128);

		/*
		 * Allocate our read/write buffer, 64KB
		 */
	if (_dos_allocmem(4096, &seg_rwbuffer))
	{
		dprintf(D_ALWAYS, "Could not allocate read/write buffer, bailing out.\n");
		goto free_all;
	}
	rwBuffer=MK_FP(seg_rwbuffer, 0);
		
		/*
		 * Allocate the (de)compression buffer 64KB
		 * we use the high 16KB for the hash table used by our
		 * compression module
		 */
	if (_dos_allocmem(4096, &seg_compbuffer))
	{
		dprintf(D_ALWAYS, "Could not allocate compression buffer, bailing out.\n");
		goto free_all;
	}
	compBuffer = MK_FP(seg_compbuffer, 0);
	lzrw1_sethash((unsigned long) compBuffer+ 0xFFFFu - LZRW1_HASHSIZE*sizeof(uint8 far *));
		
		/*
		 * Ok, if we're reach here we're in position
		 */
	tsr_print_string("Server starting up, ready for requests.", TRUE);
	debug_setout(dout);

		/*
		 * Handle requests until unload commnand from client
		 */
	handle_requests(interactive);

	tsr_print_string("Server shutdown normally.", TRUE);

		/*
		 * free all we allocated and restore drive and path
		 */
free_all:
	if (seg_rwbuffer && _dos_freemem(seg_rwbuffer))
		dprintf(D_SERVER, "Oops, clould not free read/write buffer.\n");
	if (seg_compbuffer && _dos_freemem(seg_compbuffer))
		dprintf(D_SERVER, "Oops, clould not free compression buffer.\n");
	_chdrive(old_drive);
	_chdir(old_dir);

	return 0;
}
