#pragma check_stack(off)
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include "debug.h"
#include "systypes.h"
#include "ems.h"
#include "vsprintf.h"

#define		DEBUG_STR_MAX 128		/* keep it a power of 2 */


static char 			emsDebugName[]="RMOUNT";
static union REGS 		regs;
static emshandle_t		emsDebug=0;		/* debug to ems 		*/
static unsigned 		emsPages;		/* number of ems pages 	*/
static unsigned	long	ems_pointer=0;	/* ems "filepointer"	*/

static int 				debug_level=0;
static int 				debug_out=DOUT_CONSOLE;
static char 			str_debug[DEBUG_STR_MAX];
static char 			*debug_prefix[]= 
		{
			"", 
			"System", 
			"Ems", 
			"Xdr", 
			"Server",
			"Client"
		} ;


/*
 * fill ems debug pages with eol
 */
static void debug_clearems()
{
	unsigned page,offset;
	memset(str_debug,'\n',DEBUG_STR_MAX);
	for (page=0 ; page< emsPages ; page++)
		for (offset=0 ; offset < PAGE_SIZE ; offset+=DEBUG_STR_MAX)
			emsConv2Ems(str_debug, emsDebug, page,offset, DEBUG_STR_MAX); 
}

int debug_setlevel(int newlevel)
{
	int old=debug_level;

	debug_level=newlevel;

	return old;
}

int debug_setout(int newout)
{
	int old=debug_out;

	debug_out=newout;

	return old;
}

/*
 * Initialize debug module. Set debuglevel, outputmedium (ems/console) and
 * the number of ems pages (if output to ems).
 * We always bebug to ems with a handle called 'RMOUNT', so we first
 * see if such a handle already exists, if so, we use that one (after
 * resizing the number of pages), otherwise we allocate new pages and the
 * according handle to 'RMOUNT'
 */

void debug_init(int level,int out, unsigned ems_pages)
{
	debug_level=level;
	debug_out=out;
	emsPages=ems_pages;

	if (debug_out & DOUT_EMS)
	{
		if (!(emsDebug=emsGetHandle(emsDebugName)))
			if (! (emsDebug=emsAllocPages(emsPages)))
			{
				debug_setout(DOUT_CONSOLE);
				dprintf(D_ALWAYS,"Could not alloc ems pages for debugging\n");
				return;
			}
			else
				emsSetHandleName(emsDebug,emsDebugName);
		else
		{
			emsReAllocPages(emsDebug,emsPages);
			//emsPages=emsNumPages(emsDebug);
		}
		debug_clearems();
		dprintf(D_SYSTEM,"Debug: handle:%u, Pages:%u\n",emsDebug,emsPages);
	}
}

void debug_shutdown()
{
	uint16 debugHandle;
	
/*
 * We should only clean up a handle with our debugname, this way
 * the unload routine can decide whether or not to keep the debug
 * pages.
 */

	if ((debugHandle=emsGetHandle(emsDebugName)))
		emsFreePages(debugHandle);
}

/*
 * dprintf writes a debug message to the current outputmedium (ems/console).
 * First we check the flag against the current debuglevel to see if this
 * message should be printed in the first place. Every debug class has it's
 * own prefix, except for D_ALWAYS which has none.
 */

void dprintf(int flag,char *format,...)
{
	va_list argptr;
	int which;
	unsigned len;
	static int debug_count=1;


	if ( !(((debug_level & flag) || !flag) && flag<=D_END))
		return;

	va_start(argptr, format);

	if (!flag)
		which=0;
	else
	{
		which=1;
		while (!(flag & 1))
		{
			which++;
			flag>>=1;
		}
	}

	len=0;
	if (flag)
		len=simple_sprintf(str_debug,"%-6s %5u:",debug_prefix[which],debug_count++);

	len+=simple_vsprintf(str_debug+len,format,argptr);
	va_end(argptr);
	if (!(debug_out & DOUT_NO_EOL))
	{
		strcat(str_debug,"\r");
		len++;
	}

	if (debug_out & DOUT_CONSOLE)
	{
		strcat(str_debug,"$");
		regs.x.dx=(unsigned) str_debug;
		regs.h.ah=0x09;
		int86(0x21,&regs,&regs);
	}
	else
	{
		len--;
		if (ems_pointer + len > (unsigned long) emsPages*(unsigned long) PAGE_SIZE)
			ems_pointer=0l;
		emsConv2Ems((void far *) str_debug, emsDebug, (unsigned) (ems_pointer / PAGE_SIZE), (unsigned) (ems_pointer % PAGE_SIZE), len);
		ems_pointer+=len;
	}
}


/*
 * Print the current list of debugflags known to the system
 */

void debug_showlist()
{
	int i,flag;
	for (i=flag=1 ; flag<=D_END ; i++)
	{
		dprintf(D_ALWAYS,"Flag %4u = %s\n",flag,debug_prefix[i]);
		flag<<=1;
	}
}


/*
 * This dumps a message to the console and waits for a key.
 * Useful if a bug causes a reboot or something.
 */

char press_key(char *message)
{
	int		outold;

	if (message)
	{
		outold=debug_setout(DOUT_CONSOLE);
		dprintf(D_ALWAYS,message);
		debug_setout(outold);
	}

	regs.h.ah=0;
	int86(0x16,&regs,&regs);
	return regs.h.al;
}
