#pragma check_stack(off)
#include <dos.h>

#define EMS_DECL
#include "ems.h"
#include "debug.h"
#include "vsprintf.h"

#ifndef MK_FP
#define MK_FP(a,b)  ((void far *)(((unsigned long)(a) << 16) | (b)))
#endif



static char 		emsDriverName[]="EMMXXXX0";
static char			emsVersionString[8]="0.0";
static char			emsMapBuffer[256];
static emscopy_t 	copy_buffer;
static union REGS 	emsregs;	 
static struct SREGS emssegregs;


static int 			emsGetPageFrame();
static int			emsGetSizeMapInfo();


static void	emsErrorMessage(char *func)
{
	dprintf(D_EMS,"ems error in:%s, error:%u\n",func,(unsigned) emsregs.h.ah);
}

static void emsVersion()
{

	emsregs.h.ah=0x46;
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsVersion");
	simple_sprintf(emsVersionString,"%u.%u",(unsigned) (emsregs.h.al & 0x00F0) >> 4, (unsigned) emsregs.h.al & 0x000F);
}


static void emsCopy(emscopy_t *copybuf)
{
	emsregs.h.ah=0x57;
	emsregs.h.al=0x00;
	emsregs.x.si=(unsigned)copybuf;		/* note that copybuf must be a DS:SI pointer */
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsCopy");
}

int emsPresent()
{
	_asm {


		mov  al,67h
		mov  ah,35h

		int  21h

		mov  di,10

		mov  si,offset emsDriverName
		mov  cx,8
		cld
		repz cmpsb
		jnz  error

	}
	emsVersion();
	return emsGetPageFrame();

error:
#ifdef __BORLANDC__
	emsregs.h.ah=_AH;
#else
	_asm mov emsregs.h.ah,ah;
#endif
	emsErrorMessage("emsPresent");
	return 0;
}


unsigned int emsNumFreePages()
{
	emsregs.h.ah=0x42;
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsNumFreePages");
	return emsregs.x.bx;
}

unsigned int emsNumTotPages()
{
	emsregs.h.ah=0x42;
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsNumTotPages");
	return emsregs.x.dx;
}

unsigned emsNumPages(emshandle_t Handle)
{
	emsregs.h.ah=0x4C;
	emsregs.x.dx=Handle;
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsNumPages");
	return emsregs.x.bx;
}



int emsGetPageFrame()
{

	emsregs.h.ah=0x41;
	int86(0x67,&emsregs,&emsregs);

	if (!emsregs.h.ah)
	{
		emsPageFrame=(char far *) MK_FP(emsregs.x.bx,0);
		return 1;
	}
	else
	{
		emsErrorMessage("emsGetPageFrame");
		emsPageFrame=(char far *) MK_FP(0,0);
		return 0;
	}
}


char *emsGetVersion()
{
	return emsVersionString;
}

emshandle_t emsAllocPages(unsigned int Pages)
{
	emsregs.h.ah=0x43;
	emsregs.x.bx=Pages;
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
	{
		emsErrorMessage("emsAllocPages");
		return (emshandle_t) 0;
	}
	else
		return (emshandle_t) emsregs.x.dx;
}

void emsReAllocPages(emshandle_t Handle, unsigned Pages)
{
	emsregs.h.ah=0x51;
	emsregs.x.bx=Pages;
	emsregs.x.dx=Handle;
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsReAllocPages");
}

void emsFreePages(emshandle_t Handle)
{
	emsregs.h.ah=0x45;
	emsregs.x.dx=Handle;
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsFreePages");

}

void emsSetHandleName(emshandle_t Handle, char * HandleName)
{
	emsregs.h.ah=0x53;
	emsregs.h.al=0x01;
	emsregs.x.si=(unsigned) HandleName;	/* note that HandleName must be a DS:SI pointer */
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsSetHandleName");
}

emshandle_t emsGetHandle(char * HandleName)
{
	emsregs.h.ah=0x54;
	emsregs.h.al=0x01;
	emsregs.x.si=(unsigned) HandleName;	/* note that HandleName must be a DS:SI pointer */
	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsSetHandle");

	return emsregs.x.dx;
}

int emsMapPage(emshandle_t Handle, char pPage, unsigned lPage)
{
	emsregs.h.ah=0x44;
	emsregs.h.al=pPage;
	emsregs.x.bx=lPage;
	emsregs.x.dx=Handle;

	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
	{
		emsErrorMessage("emsMapPage");
		return 0;
	}
	else
		return 1;
}

void emsSaveMap()
{
	dprintf(D_EMS,"SaveMap: %i bytes are needed\n",emsGetSizeMapInfo());

	emsregs.h.ah=0x4e;
	emsregs.h.al=0x00;
	emsregs.x.di=(unsigned) emsMapBuffer;
#ifdef __BORLANDC__
	emssegregs.es=_DS;
#else
	_asm mov emssegregs.es, ds
#endif

	int86x(0x67,&emsregs,&emsregs,&emssegregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsSaveMap");
}
void emsRestoreMap()
{
	emsregs.h.ah=0x4e;
	emsregs.h.al=0x01;
	emsregs.x.si=(unsigned) emsMapBuffer;

	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsRestoreMap");
}

int emsGetSizeMapInfo()
{

	emsregs.h.ah=0x4E;
	emsregs.h.al=0x03;

	int86(0x67,&emsregs,&emsregs);
	if (emsregs.h.ah)
		emsErrorMessage("emsGetSizeMapMap");

	return (int) emsregs.h.al;
}


void emsConv2Ems(void far *src, emshandle_t emsHandle, unsigned Page, unsigned offset, unsigned long size)
{
	unsigned long p_addr= ((((unsigned long) FP_SEG(src)) >> 4) << 8) 
							+ (unsigned long )FP_OFF(src);

	copy_buffer.len=size;
	copy_buffer.src_type='\0';
	copy_buffer.src_offset=FP_OFF(src);
	copy_buffer.src_seg_or_page=FP_SEG(src);

	copy_buffer.dst_type='\1';
	copy_buffer.dst_handle=emsHandle;
	copy_buffer.dst_offset=offset;
	copy_buffer.dst_seg_or_page=Page;
	emsCopy(&copy_buffer);
}

void emsEms2Conv(emshandle_t emsHandle, unsigned Page, unsigned offset, void far *dst,unsigned long size)
{
	copy_buffer.len=size;
	copy_buffer.src_type='\1';
	copy_buffer.src_handle=emsHandle;
	copy_buffer.src_offset=offset;
	copy_buffer.src_seg_or_page=Page;
	copy_buffer.dst_type='\0';
	copy_buffer.dst_offset=FP_OFF(dst);
	copy_buffer.dst_seg_or_page=FP_SEG(dst);
	emsCopy(&copy_buffer);

}

void emsEms2Ems(emshandle_t src_emsHandle, unsigned src_Page, unsigned src_offset, emshandle_t dst_emsHandle, unsigned dst_Page, unsigned dst_offset, unsigned long size)
{
	copy_buffer.len=size;
	copy_buffer.src_type='\1';
	copy_buffer.src_handle=src_emsHandle;
	copy_buffer.src_offset=src_offset;
	copy_buffer.src_seg_or_page=src_Page;
	copy_buffer.dst_type='\1';
	copy_buffer.dst_handle=dst_emsHandle;
	copy_buffer.dst_offset=dst_offset;
	copy_buffer.dst_seg_or_page=dst_Page;
	emsCopy(&copy_buffer);

}
