#ifndef __EMS_H
#define __EMS_H


/*
 * This implements the basic ems stuff for our driver. Most of
 * it are 3.0 calls, only the copy functions are 4.0
 */

#define PAGE_SIZE 16384

typedef  unsigned int emshandle_t;

#pragma pack(1)
typedef  struct
{
	unsigned long		len;
	char				src_type;
	emshandle_t			src_handle;
	unsigned			src_offset;
	unsigned			src_seg_or_page;
	char				dst_type;
	emshandle_t			dst_handle;
	unsigned			dst_offset;
	unsigned			dst_seg_or_page;
} emscopy_t;
#pragma pack()

int 		emsPresent();


/*
 * Info routines
 *
 */

unsigned 	emsNumFreePages();
unsigned 	emsNumTotPages();
unsigned	emsNumPages(emshandle_t);
char* 		emsGetVersion();


/*
 * Alloc/free
 *
 */

emshandle_t emsAllocPages(unsigned int Pages);
void 		emsReAllocPages(emshandle_t Handle,unsigned int Pages);
void 		emsFreePages(emshandle_t Handle);

/*
 *  Handle Names
 */

void emsSetHandleName(emshandle_t Handle, char * HandleName);
emshandle_t emsGetHandle(char * HandleName);

/*
 * Map a Page (pPage=[0..3] ; lPage=[0..n-1], n=number of alloced pages)
 *
 */

int 		emsMapPage(emshandle_t Handle, char pPage, unsigned lPage);


/*
 * Save/restore context
 *
 */
void 		emsSaveMap();
void 		emsRestoreMap();

/*
 * Some copy routines, these are 4.0, they do not touch the context mapping.
 * Note the far pointers.
 */

void emsConv2Ems(void far *src, emshandle_t emsHandle, unsigned Page, unsigned offset, unsigned long size);
void emsEms2Conv(emshandle_t emsHandle, unsigned Page, unsigned offset, void far *dest,unsigned long size);
void emsEms2Ems(emshandle_t src_emsHandle, unsigned src_Page, unsigned src_offset, emshandle_t dst_emsHandle, unsigned dst_Page, unsigned dst_offset, unsigned long size);


#ifdef EMS_DECL
char far * emsPageFrame;
#else
extern char far * emsPageFrame;
#endif
#undef EMS_DECL

#endif		// __EMS_H 
